#include "Camera.h"

#include <math.h>
#include <cassert>

#ifndef M_PI
#define M_PI (3.1415926536)
#endif

using namespace ambenv;

Camera::Camera(const std::string & name) : m_left( -1.0 ), m_right( 1.0 ), 
					   m_bottom( -1.0 ), m_top( 1.0 ), 
					   m_near( 1.0 ), m_far( 1000.0 ), 
					   m_center( 0, 0, -1 ), m_up( 0, 1, 0), m_aspect( 1.3333 )
{
  Nameable::setName( name );
}

Camera::Camera()
{
  Camera( "" );
}

/*!
 * All parameters should be positive, fov in [0, 180].
 */
void Camera::perspective( Real fov, Real aspect, Real vnear, Real vfar )
{
  assert( fov > 0.0 );
  assert( fov < 180.0 );
  assert( aspect > 0.0 );
  assert( vnear > 0.0 );
  assert( vfar > vnear );

  Real vfov = atan( tan ( fov * M_PI / 360 ) / aspect );
  
  Real ymax = vnear * (Real)tan( vfov );
  Real xmax = ymax * aspect;
  
  frustum( -xmax, xmax, -ymax, ymax, vnear, vfar );
}

/*!
 * See the OpenGL redbook, Appendix F, for reference.
 * Valid only if (left != right) and (top != bottom) and (near != far)
 */
void Camera::frustum( Real left, Real right,
		      Real bottom, Real top,
		      Real vnear, Real vfar )
{
  // -- validate the parameters
  assert( vnear != vfar );
  assert( right != left );
  assert( top != bottom );
  assert( vnear > 0.0 );
  assert( vfar > vnear );

  m_left = left;
  m_right = right;
  m_bottom = bottom;
  m_top = top;
  m_near = vnear;
  m_far = vfar;

  m_aspect = (right - left) / (top - bottom);

  // -- find the matrix
  Real x = (2.0 * vnear) / (right - left);
  Real y = (2.0 * vnear) / (top - bottom);
  Real a = (right + left) / (right - left);
  Real b = (top + bottom) / (top - bottom);
  Real c = (vfar + vnear) / (vnear - vfar);
  Real d = (2.0 * vfar * vnear) / (vnear - vfar);

  m_persp.identity();
  m_persp( 0, 0, x );
  m_persp( 1, 1, y );
  m_persp( 0, 2, a );
  m_persp( 1, 2, b );
  m_persp( 2, 2, c );
  m_persp( 2, 3, d );
  m_persp( 3, 2, -1.0 );
  m_persp( 3, 3, 0.0 );

  // -- now find the inverse
  x = (right - left) / (2.0 * vnear);
  y = (top - bottom) / (2.0 * vnear);
  a = -(vfar - vnear) / (2.0 * vfar * vnear);
  b = (right + left) / (2.0 * vnear);
  c = (top + bottom) / (2.0 * vnear);
  d = (vfar + vnear) / (2.0 * vfar * vnear);

  m_inv_persp.zero();
  m_inv_persp( 0, 0, x );
  m_inv_persp( 1, 1, y );
  m_inv_persp( 3, 2, a );
  m_inv_persp( 0, 3, b );
  m_inv_persp( 1, 3, c );
  m_inv_persp( 2, 3, -1.0 );
  m_inv_persp( 3, 3, d );
}

/*!
 * See the OpenGL redbook, Appendix F, for reference.
 * Valid only if (left != right) and (top != bottom) and (near != far)
 */
void Camera::ortho( Real left, Real right, Real bottom, Real top, Real vnear, Real vfar )
{
  // -- validate the parameters
  assert( vnear != vfar );
  assert( right != left );
  assert( top != bottom );
  assert( vfar != 0.0 );
  assert( vnear != 0.0 );


  // -- find the matrix
  Real x = 2.0 / (right - left);
  Real y = 2.0 / (top - bottom);
  Real z = -2.0 / (vfar - vnear);
  Real a = (right + left) / (right - left);
  Real b = (top + bottom) / (top - bottom);
  Real c = (vfar + vnear) / (vfar - vnear);

  m_persp.zero();
  m_persp( 0, 0, x );
  m_persp( 1, 1, y );
  m_persp( 2, 2, z );
  m_persp( 0, 3, a );
  m_persp( 1, 3, b );
  m_persp( 2, 3, c );
  m_persp( 3, 3, 1.0 );


  // -- now find the inverse
  x = (right - left) / 2.0;
  y = (top - bottom) / 2.0;
  z = (vfar - vnear) / -2.0;
  a = (right + left) / 2.0;
  b = (top + bottom) / 2.0;
  c = (vnear + vfar) / 2.0;
  
  m_inv_persp.zero();
  m_inv_persp( 0, 0, x );
  m_inv_persp( 1, 1, y );
  m_inv_persp( 2, 2, z );
  m_inv_persp( 0, 3, a );
  m_inv_persp( 1, 3, b );
  m_inv_persp( 2, 3, c );
  m_inv_persp( 3, 3, 1.0 );
}

void Camera::lookAt( const Vec3 & eye, const Vec3 & center, const Vec3 & up )
{
  m_eye = eye;
  m_center = center;
  m_up = up;

  Vec3 w = eye - center;
  Real mag = w.length();
  if( mag != 0.0 ) {
    w /= mag;
  }

  Vec3 u = up.cross( w );
  mag = u.length();
  if( mag != 0.0 ) {
    u /= mag;
  }

  Vec3 v = w.cross( u );
  mag = v.length();
  if( mag != 0.0 ) {
    v /= mag;
  }

  // -- create the rotation matrix, defined by the u,v,w coordinate system
  Matrix4x4 R;
  R( 0, 0, u.x() );
  R( 0, 1, u.y() );
  R( 0, 2, u.z() );
  R( 1, 0, v.x() );
  R( 1, 1, v.y() );
  R( 1, 2, v.z() );
  R( 2, 0, w.x() );
  R( 2, 1, w.y() );
  R( 2, 2, w.z() );
  
  // -- create the inverse rotation, which is just the transpose
  Matrix4x4 Rinv = R.transpose();

  // -- create translation matrix
  Matrix4x4 T;
  T( 0, 3, -eye.x() );
  T( 1, 3, -eye.y() );
  T( 2, 3, -eye.z() );

  // -- create the inverse translation matrix, which is just translation in
  // -- the opposite direction, of course, of course
  Matrix4x4 Tinv;
  Tinv( 0, 3, eye.x() );
  Tinv( 1, 3, eye.y() );
  Tinv( 2, 3, eye.z() );

  m_view = R * T;
  m_inv_view =  Tinv * Rinv;
}

Matrix4x4 Camera::getPersp() const
{
  return( m_persp );
}

void Camera::setPersp( const Matrix4x4 & mat )
{
  m_persp = mat;
}

Matrix4x4 Camera::getInversePersp() const
{
  return( m_inv_persp );
}

void Camera::setInversePersp( const Matrix4x4 & mat )
{
  m_inv_persp = mat;
}

Matrix4x4 Camera::getView() const
{
  return( m_view );
}

void Camera::setView( const Matrix4x4 & mat )
{
  m_view = mat;
}

Matrix4x4 Camera::getInverseView() const
{
  return( m_inv_view );
}

void Camera::setInverseView( const Matrix4x4 & mat )
{
  m_inv_view = mat;
}















