#include "Matrix4x4.h"

#include <stdlib.h>
#include <string.h>
#include <cassert>

#include "gaussj.h"

using namespace ambenv;

Matrix4x4::Matrix4x4()
{
  identity();
}

Matrix4x4::Matrix4x4( const Matrix4x4 & rhs )
{
  for(unsigned i=0; i<M_DIM; i++) {
    for(unsigned j=0; j<M_DIM; j++) {
      (*this)(i, j, rhs(i, j));
    }
  }
}

void Matrix4x4::zero()
{
  memset( mp_data, 0, M_DIM*M_DIM*sizeof( Real ) );
}

void Matrix4x4::identity()
{
  zero();
  for(unsigned i=0; i<M_DIM; i++) {
    for(unsigned j=0; j<M_DIM; j++) {
      if(i == j) {
	(*this)( i, j, 1.0 );
      }
    }
  }
}

bool Matrix4x4::operator==( const Matrix4x4 & rhs ) const
{
  for(unsigned i=0; i<M_DIM; i++) {
    for(unsigned j=0; j<M_DIM; j++) {
      if( (*this)( i, j ) != rhs( i, j ) ) {
	return( false );
      }
    }
  }
  return( true );
}

bool Matrix4x4::operator!=( const Matrix4x4 & rhs ) const
{
  return( !this->operator==( rhs ) );
}

Matrix4x4 Matrix4x4::operator*( const Matrix4x4 & rhs ) const
{
  Matrix4x4 result;
  for(unsigned i=0; i<M_DIM; i++) {
    for(unsigned j=0; j<M_DIM; j++) {
      Real value = 0.0;
      for(unsigned k=0; k<M_DIM; k++) {
	value += ( (*this)(i, k) * rhs(k, j) );
      }
      result(i, j, value);
    }
  }
  return( result );
}

Matrix4x4 Matrix4x4::operator*( Real value ) const
{
  Matrix4x4 result;
  for(unsigned i=0; i<M_DIM; i++) {
    for(unsigned j=0; j<M_DIM; j++) {
      result( i, j, (*this)( i, j) * value );
    }
  }
  return( result );
}

Matrix4x4 Matrix4x4::operator/( Real value ) const
{
  assert( value != 0.0 );
  return( (*this) * (1.0 / value) );
}

Matrix4x4 Matrix4x4::operator+( const Matrix4x4 & rhs ) const
{
  Matrix4x4 result;
  for(unsigned i=0; i<M_DIM; i++) {
    for(unsigned j=0; j<M_DIM; j++) {
      Real value = ( (*this)(i, j) + rhs(i, j) );
      result(i, j, value);
    }
  }
  return( result );
}

Matrix4x4 Matrix4x4::operator-( const Matrix4x4 & rhs ) const
{
  Matrix4x4 result;;
  for(unsigned i=0; i<M_DIM; i++) {
    for(unsigned j=0; j<M_DIM; j++) {
      Real value = ( (*this)(i, j) - rhs(i, j) );
      result(i, j, value);
    }
  }
  return( result );
}

void Matrix4x4::operator*=( const Matrix4x4 & rhs )
{
  (*this) = (*this) * rhs;
}

void Matrix4x4::operator*=( Real value )
{
  for(unsigned i=0; i<M_DIM; i++) {
    for(unsigned j=0; j<M_DIM; j++) {
      (*this)( i, j, (*this)( i, j) * value );
    }
  }
}

void Matrix4x4::operator/=( Real value )
{
  assert( value != 0.0 );
  (*this) *= (1.0 / value);
}

void Matrix4x4::operator+=( const Matrix4x4 & rhs )
{
  for(unsigned i=0; i<M_DIM; i++) {
    for(unsigned j=0; j<M_DIM; j++) {
       (*this)( i, j, (*this)( i, j ) + rhs(i, j) );
    }
  }
}

void Matrix4x4::operator-=( const Matrix4x4 & rhs ) 
{
  for(unsigned i=0; i<M_DIM; i++) {
    for(unsigned j=0; j<M_DIM; j++) {
      (*this)( i, j, (*this)( i, j ) + rhs(i, j) );
    }
  }
}

Vec3 Matrix4x4::multVec( const Vec3 & v, bool homogeneous ) const
{
  Vec3 result( v.x() * mp_data[0] + v.y() * mp_data[4] + v.z() * mp_data[8],
	       v.x() * mp_data[1] + v.y() * mp_data[5] + v.z() * mp_data[9],
	       v.x() * mp_data[2] + v.y() * mp_data[6] + v.z() * mp_data[10] );
  
  if( homogeneous ) {
    result.x( result.x() + mp_data[12] );
    result.y( result.y() + mp_data[13] );
    result.z( result.z() + mp_data[14] );
  }
  
  return( result );
}

Real Matrix4x4::operator()( unsigned i, unsigned j ) const
{
  assert( i < M_DIM );
  assert( j < M_DIM );
    
  return( mp_data[M_DIM*j + i] );
}

void Matrix4x4::operator()( unsigned i, unsigned j, Real value )
{
  assert( i < M_DIM );
  assert( j < M_DIM );

  mp_data[M_DIM*j + i] = value;
}

Matrix4x4 Matrix4x4::transpose() const
{
  Matrix4x4 result;

  for(unsigned i=0; i<M_DIM; i++) {
    for(unsigned j=0; j<M_DIM; j++) {
      result( j, i, (*this)( i, j ) );
    }
  }
  
  return( result );
}

bool Matrix4x4::invert()
{
  unsigned i, j;
  float ** a = new float *[M_DIM+1];
  float ** b = new float * [M_DIM+1];
  for(i=1; i<=M_DIM; i++) {
    a[i] = new float[M_DIM+1];
    b[i] = new float[2];
    b[i][0] = 0.0f;
    b[i][1] = 0.0f;
  }

  for(i=1; i<=M_DIM; i++) {
    for(j=1; j<=M_DIM; j++) {
      a[i][j] = (float)(*this)( (i-1), (j-1) );
    }
  }

  bool retval = true;
  if( gaussj( a, M_DIM, b, 1 ) != 1 ) {
    retval = false;
  } else {
    for(i=1; i<=M_DIM; i++) {
      for(j=1; j<=M_DIM; j++) {
	(*this)( (i-1), (j-1), (Real)a[i][j] );
      }
    }
  }
  
  for(i=1; i<=M_DIM; i++) {
    delete [] a[i];
    delete [] b[i];
  }
  delete [] a;
  delete [] b;

  return( retval );
}

















