#include "Scene.h"
#include "util.h"

#include <lib3ds/file.h>
#include <lib3ds/vector.h>
#include <lib3ds/matrix.h>
#include <lib3ds/camera.h>
#include <lib3ds/light.h>
#include <lib3ds/material.h>
#include <lib3ds/mesh.h>
#include <lib3ds/node.h>

#include <string>

// -- the string name for a 3DS null node
const std::string NULL_3DS_NAME = "$$$DUMMY";

using namespace ambenv;

Scene::Scene() : width( 800 ), height( 600 ), bpp( 24 ),
		 fullscreen( false ), samples( 5 ), usehemicube( false )
{
}

Scene::~Scene()
{
  clear();
}

void Scene::clear()
{
  {
    std::list<Camera *>::iterator itr = cameras.begin();
    for( ; itr != cameras.end(); ++itr) {
      if( *itr ) {
	delete *itr;
	*itr = NULL;
      }
    }
    cameras.clear();
  }
  {
    std::list<Light *>::iterator itr = lights.begin();
    for( ; itr != lights.end(); ++itr) {
      if( *itr ) {
        delete *itr;
        *itr = NULL;
      }
    }
    lights.clear();
  }
  {
    std::list<Instance *>::iterator itr = instances.begin();
    for( ; itr != instances.end(); ++itr) {
      if( *itr ) {
        delete *itr;
        *itr = NULL;
      }
    }
    instances.clear();
  }
  {
    std::map<std::string, Object *>::iterator itr = objects.begin();
    for( ; itr != objects.end(); ++itr) {
      if( itr->second ) {
        delete itr->second;
        itr->second = NULL;
      }
    }
    objects.clear();
  }
  
  {
    std::map<std::string, Material *>::iterator itr = materials.begin();
    for( ; itr != materials.end(); ++itr) {
      if( itr->second ) {
        delete itr->second;
        itr->second = NULL;
      }
    }
    materials.clear();
  }
  
}

void Scene::getBoundingBox( Vec3 box[8] )
{
  
  box[0] = Vec3( REAL_MAX, REAL_MAX, -REAL_MAX );
  box[1] = Vec3( REAL_MAX, -REAL_MAX, -REAL_MAX );
  box[2] = Vec3( -REAL_MAX, -REAL_MAX, -REAL_MAX );
  box[3] = Vec3( -REAL_MAX, REAL_MAX, -REAL_MAX );
  box[4] = Vec3( REAL_MAX, REAL_MAX, REAL_MAX );
  box[5] = Vec3( REAL_MAX, -REAL_MAX, REAL_MAX );
  box[6] = Vec3( -REAL_MAX, -REAL_MAX, REAL_MAX );
  box[7] = Vec3( -REAL_MAX, REAL_MAX, REAL_MAX );

  std::list<Instance *>::iterator itr = instances.begin();
  for( ; itr != instances.end(); ++itr) {

    const Mesh * mesh = (Mesh *)((*itr)->object());
    assert( mesh != NULL );

    for( unsigned i = 0; i < mesh->numVertices(); i++ ) {
      Vec3 v = mesh->getVertex(i).v();

      if( v.x() < box[0].x() ) { box[0].x(v.x()); }
      if( v.y() < box[0].y() ) { box[0].y(v.y()); }
      if( v.z() > box[0].z() ) { box[0].z(v.z()); }

      if( v.x() < box[1].x() ) { box[1].x(v.x()); }
      if( v.y() > box[1].y() ) { box[1].y(v.y()); }
      if( v.z() > box[1].z() ) { box[1].z(v.z()); }
      
      if( v.x() > box[2].x() ) { box[2].x(v.x()); }
      if( v.y() > box[2].y() ) { box[2].y(v.y()); }
      if( v.z() > box[2].z() ) { box[2].z(v.z()); }

      if( v.x() > box[3].x() ) { box[3].x(v.x()); }
      if( v.y() < box[3].y() ) { box[3].y(v.y()); }
      if( v.z() > box[3].z() ) { box[3].z(v.z()); }
    
      if( v.x() < box[4].x() ) { box[4].x(v.x()); }
      if( v.y() < box[4].y() ) { box[4].y(v.y()); }
      if( v.z() < box[4].z() ) { box[4].z(v.z()); }

      if( v.x() < box[5].x() ) { box[5].x(v.x()); }
      if( v.y() > box[5].y() ) { box[5].y(v.y()); }
      if( v.z() < box[5].z() ) { box[5].z(v.z()); }

      if( v.x() > box[6].x() ) { box[6].x(v.x()); }
      if( v.y() > box[6].y() ) { box[6].y(v.y()); }
      if( v.z() < box[6].z() ) { box[6].z(v.z()); }

      if( v.x() > box[7].x() ) { box[7].x(v.x()); }
      if( v.y() < box[7].y() ) { box[7].y(v.y()); }
      if( v.z() < box[7].z() ) { box[7].z(v.z()); }
    } 
  }

  return;
}

bool Scene::load3DS( const std::string & filename )
{
  // -- load the 3DS file
  Lib3dsFile * fp_3ds = lib3ds_file_load( filename.c_str() );
  if( !fp_3ds ) {
    return( false );
  }
	
  lib3ds_file_eval( fp_3ds, 0 );

  std::string str = 
    filename.substr( filename.find_last_of( '/' )+1, filename.size() );
  
  if( str.empty() ) {
    str = "default";
  }

  Mesh * mnode = new Mesh( str );

  {
    Lib3dsNode * p;
    for( p = fp_3ds->nodes; p != 0; p = p->next ) {
      load3dsNode( fp_3ds, p, mnode );
    }
  }

  generate_vertex_normals( *mnode );

  objects.insert( std::pair<std::string, Object *>( mnode->getName(), mnode ) );
  instances.push_back( new Instance( mnode->getName(), "", mnode ) );

  // -- clean up 3DS file
  lib3ds_file_free( fp_3ds );

  if( !load3dsTextures( filename ) ) {
    return( false );
  }
  
  return( true );
}

bool Scene::load3dsNode( void * vp_file, void * vp_node, Mesh * mnode )
{
  assert( vp_file != NULL );
  assert( vp_node != NULL );

  Lib3dsFile * fp_3ds = (Lib3dsFile *)vp_file;
  Lib3dsNode * node = (Lib3dsNode *)vp_node;
 
  // -- Below here is the meat of the mesh loading code.
  /*! This was adapted from the "3ds2rib" example. See the
   *  lib3ds website for more documentation. (lib3ds.sourceforge.net)
   */
  Lib3dsMesh * mesh = NULL;
  
  if( (node->type == LIB3DS_OBJECT_NODE) && 
      (std::string( node->name ) != NULL_3DS_NAME ) ) {
    
    // -- get the 3ds mesh structure
    mesh = lib3ds_file_mesh_by_name( fp_3ds, node->name );    
    if( !mesh ) {
      return( false );
    }

    int vertexOffset = mnode->numVertices();

    //Lib3dsObjectData * d = &node->data.object;

    // -- Rotate Z up axis into Y up axis.
    // -- Or, like rotate( -90, 1, 0, 0 )
    Matrix4x4 xform;
    xform( 1, 1, 0.0 );
    xform( 1, 2, 1.0 );
    xform( 2, 1, -1.0 );
    xform( 2, 2, 0.0 );

    {
      /*
      Lib3dsMatrix N, M, X;
      lib3ds_matrix_copy( N, node->matrix );
      lib3ds_matrix_translate_xyz( N, -d->pivot[0], -d->pivot[1], -d->pivot[2] );
      lib3ds_matrix_rotate_x( N, -0.5 * REAL_PI );
      lib3ds_matrix_copy( M, mesh->matrix );
      lib3ds_matrix_inv( M );
      lib3ds_matrix_mul( X, N, M );

      
      Lib3dsMatrix X;
      lib3ds_matrix_identity( X );
      lib3ds_matrix_rotate_x( X, -0.5 * REAL_PI );
      

      // -- set this transform for this node
      for( unsigned i=0; i<M_DIM; i++ ) {
	for( unsigned j=0; j<M_DIM; j++ ) {
	  xform( i, j, (Real)X[j][i] );
	}
      }
      */
      //std::cout << xform;
      ///inv_xform = xform;
      //bool result = inv_xform.invert();
      //assert( result );
    }


    {   
      for( unsigned p=0; p<mesh->faces; ++p ) {
	Lib3dsFace * face = &mesh->faceL[p];
	Lib3dsMaterial * mat = lib3ds_file_material_by_name( fp_3ds, face->material );

	if( mat ) {
	  std::string name( mat->name );

	  if( materials.find( name ) == materials.end() ) {

	    Vec3 diffuse( mat->diffuse[0],
			  mat->diffuse[1],
			  mat->diffuse[2] );
	    
	    Vec3 ambient( mat->ambient[0],
			  mat->ambient[1],
			  mat->ambient[2] );
	    
	    Vec3 specular( mat->specular[0],
			   mat->specular[1],
			   mat->specular[2] );
	    
	    Real spec_power = mat->shininess;
	  
	    Material * material = new Material( name, Vec3(), ambient, diffuse, 
						specular, spec_power,
						mat->texture1_map.name );

	    materials.insert( std::pair<std::string, Material *>
			      ( name, material ) );

	    material->materialId( materials.size() );
	  }

	  if( strlen( mat->texture1_map.name ) > 0 ) {
	    
	    std::string name( mat->texture1_map.name );

	    if( textures.find( name ) == textures.end() ) {
	      Texture * tp = new Texture();
	      textures.insert( std::pair<std::string, Texture *>( name, tp ) );
	      tp->m_textureId = textures.size();
	    }
	  }

	  
	  if( strlen( mat->texture2_map.name ) > 0 ) {

	    std::string name( mat->texture2_map.name );
	    
	    if( textures.find( name ) == textures.end() ) {
	      Texture * tp = new Texture();
	      textures.insert( std::pair<std::string, Texture *>( name, tp ) );
	      tp->m_textureId = textures.size();
	    }
	  } 
	}

      }

      for( unsigned i=0; i<mesh->faces; i++ ) {

	MeshTriangle triangle;

	Lib3dsMaterial * mat = lib3ds_file_material_by_name( fp_3ds, mesh->faceL[i].material );	
	if( mat != NULL ) {
	  
	  if( strlen( mat->texture1_map.name ) > 0 ) {
	    std::string name = mat->texture1_map.name;
	    
	    std::map<std::string, Texture *>::const_iterator t_itr =
	      textures.find( name );

	    if( t_itr != textures.end() ) {
	      triangle.textureId( t_itr->second->m_textureId );
	    } else {
	      triangle.textureId( 0 );
	    }
	  }


	  std::string name = mat->name;
	  std::map<std::string, Material *>::const_iterator m_itr =
	    materials.find( name );
	  
	  if( m_itr != materials.end() ) {
	    triangle.materialId( m_itr->second->materialId() );
	  } else {
	    triangle.materialId( 0 );
	  }
	}

	for( unsigned j=0; j<3; j++ ) {
	  triangle.vert( j, mesh->faceL[i].points[j] + vertexOffset );
	}

	mnode->addTriangle( triangle );

	MeshTriangle t = mnode->getTriangle( mnode->numTriangles() - 1 );
	if( t.materialId() != triangle.materialId() ) {
	  std::cout << "What the hell?\n";
	}

      }


      Vec3 default_normal( 0.0, 1.0, 0.0 );

      if( mesh->texelL && (mesh->texels == mesh->points) ) {
	for( unsigned i=0; i<mesh->points; i++ ) {
	  Vec3 vertex( (Real)mesh->pointL[i].pos[0],
		       (Real)mesh->pointL[i].pos[1],
		       (Real)mesh->pointL[i].pos[2] );

	  Vec3 tcoord( (Real)mesh->texelL[i][0],
		       (Real)mesh->texelL[i][1],
		       0.0 );

	  mnode->addVertex( xform.multVec( vertex ), 
			    default_normal, 
			    tcoord );
	}
      } else {
	for( unsigned i=0; i<mesh->points; i++ ) {
	  Vec3 vertex( (Real)mesh->pointL[i].pos[0],
		       (Real)mesh->pointL[i].pos[1],
		       (Real)mesh->pointL[i].pos[2] );
	  
	  mnode->addVertex( xform.multVec( vertex ), 
			    default_normal, 
			    Vec3() );
	}
      }    
    }

  } /* end if object node */

  bool retval = true;
  {
    Lib3dsNode * p;
    for( p=node->childs; p != 0; p = p->next ) {
      if( !load3dsNode( fp_3ds, p, mnode ) ) {
	retval = false;
      }
    }
  }

  return( retval );
}

bool Scene::load3dsTextures( const std::string & filename )
{

  std::string filepath = get_path( filename ) + "/maps/";

  bool retval = true;

  std::map<std::string, Texture *>::iterator itr = textures.begin();
  for( ; itr != textures.end(); ++itr ) {

    std::string rel_filename = filepath + itr->first;
    Texture * pt = itr->second;
    
    if( !read_rgba_tiff( rel_filename, pt->m_width, pt->m_height, pt->m_data ) ) {
      retval = false;
    }
  }

  return( retval );
}
