#include "util.h"

#include <SDL/SDL_opengl.h>
#include <tiffio.h>

using namespace ambenv;

// -- stream print functions
std::ostream & ambenv::operator<<( std::ostream & output, const Vec3 & vec )
{
  output << "v = [" << vec.x() << " " << vec.y() << " " << vec.z() << "]\n";
  return( output );
}

std::ostream & ambenv::operator<<( std::ostream & output, const Matrix4x4 & mat )
{
  for(unsigned i=0; i<4; i++) {
    for(unsigned j=0; j<4; j++) {
      output << mat(i, j) << " ";
    }
    output << "\n";
  }
  output << "\n";
  return( output );
}

std::ostream & ambenv::operator<<( std::ostream & output, const Ray & ray )
{
  Vec3 o = ray.getOrigin();
  Vec3 d = ray.getDirection();
  output << "r(t) = [" << o.x() << " " << o.y() << " " << o.z() << "] + t["
	 << d.x() << " " << d.y() << " " << d.z() << "]\n";
  return( output );
}

std::ostream & ambenv::operator<<( std::ostream & output, const Scene & scene )
{
  output << "Scene:\n";
  output << "   Objects:         " << scene.objects.size() << "\n";
  std::map<std::string, Object *>::const_iterator c_itr = scene.objects.begin();
  for( ; c_itr != scene.objects.end(); ++c_itr ) {
    output << "      " << ((const Mesh *)(c_itr->second)) << "\n";
  }
  output << "   Materials:       " << scene.materials.size() << "\n";
  output << "   Instances:       " << scene.instances.size() << "\n";
  output << "   Lights:          " << scene.lights.size() << "\n";
  output << "   Cameras:         " << scene.cameras.size() << "\n";
  output << "   Textures:        " << scene.textures.size() << "\n";

  /*   
  std::map<std::string, Texture *>::const_iterator i = scene.textures.begin();
  for( ; i != scene.textures.end(); ++i ) {
    output << "       " << i->first << " " << i->second->m_textureId << "\n";
  }
  */

  return( output );
}

std::ostream & ambenv::operator<<( std::ostream & output, const Mesh * mesh )
{
  output << mesh->numTriangles() << " triangles  ";
  output << mesh->numVertices() << " vertices";
  
  return( output );
}

std::string ambenv::get_path( const std::string filename )
{
  return( filename.substr( 0, filename.find_last_of( '/' ) ) );
}


bool ambenv::write_rgb_tiff( const std::string & filename, int w, int h, unsigned char * data )
{
  TIFF * file = TIFFOpen( filename.c_str(), "w" );
  if( NULL != file ) {
    TIFFSetField( file, TIFFTAG_IMAGEWIDTH, (uint32)w );
    TIFFSetField( file, TIFFTAG_IMAGELENGTH, (uint32)h );
    TIFFSetField( file, TIFFTAG_BITSPERSAMPLE, 8 );
    TIFFSetField( file, TIFFTAG_COMPRESSION,  COMPRESSION_NONE );
    TIFFSetField( file, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB );
    TIFFSetField( file, TIFFTAG_SAMPLESPERPIXEL, 3 );
    TIFFSetField( file, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG );
    TIFFSetField( file, TIFFTAG_ROWSPERSTRIP, 1 );
    TIFFSetField( file, TIFFTAG_IMAGEDESCRIPTION, "rt-prad output" );

    unsigned char * p = data;
    for (int i = ((int)h - 1); i >= 0; i--) {
      if( TIFFWriteScanline( file, p, i, 0 ) < 0 ) {
        break;
      }
      p += w * sizeof(unsigned char) * 3;
    }
    TIFFClose( file );

    std::cout << "Wrote file:         " << filename << "\n";

    return( true );
  }
  return( false );
}

bool ambenv::read_rgba_tiff( const std::string & filename, unsigned & w, unsigned & h,
				      std::vector<unsigned char> & buffer )
{
  buffer.clear();
  TIFF* tif = TIFFOpen( filename.c_str(), "r" ); 
  if( tif ) { 
    TIFFGetField( tif, TIFFTAG_IMAGEWIDTH, &w ); 
    TIFFGetField( tif, TIFFTAG_IMAGELENGTH, &h ); 
    size_t npixels = (w * h);
    if( npixels > 0 ) {
      uint32 * raster = (uint32 *) _TIFFmalloc( npixels * sizeof( uint32 ) ); 
      if( raster != NULL ) { 
	if( TIFFReadRGBAImage( tif, w, h, raster, 0 ) ) {
	  buffer.resize( npixels * 4 );
	  
	  // -- Copy the ABGR packed into integers in the TIFF buffer
	  // -- into the RGBA uchar buffer.
	  unsigned char * p = &buffer[0];
	  for( unsigned i=0; i<h; i++ ) {
	    for( unsigned j=0; j<w; j++) {
	      for( unsigned k=0; k<4; k++ ) {
		*p = (raster[i*w + j] >> k*8) & 0xff;
		p++;
	      }
	    }
	  }

	}
      } 
      _TIFFfree( raster ); 
    }
    TIFFClose( tif );

    return( true );
  }
    return( false );
}

unsigned ambenv::power_of_two( unsigned input )
{
  unsigned value = 1;
  while ( value < input ) {
    value <<= 1;
  }
  return( value );
}

unsigned char ambenv::convert_to_uchar( Real v )
{
  v *= 255.0;
  if( v < 0.0 ) {
    return( 0 );
  } else if( v > 255.0 ) {
    return( 255 );
  } else {
    return( (unsigned char)v );
  }
}

void ambenv::copy_to_farray( const Vec3 & v, float a[3] )
{
  a[0] = (float)v.x();
  a[1] = (float)v.y();
  a[2] = (float)v.z();
}

Real ambenv::get_tri_area( const Vec3 vertex[3], const Vec3 & normal )
{
  Real area = 0.0;
  Real ax, ay, az;

  // -- choose which coordinate to ignore, pick the largest in absolute value
  ax = (normal.x() > 0 ? normal.x() : -normal.x());     // abs x-coord
  ay = (normal.y() > 0 ? normal.y() : -normal.y());     // abs y-coord
  az = (normal.z() > 0 ? normal.z() : -normal.z());     // abs z-coord

  int coord = 2;
  if( ax > ay ) {
    if( ax > az ) {
      coord = 0;
    }
  } else if( ay > az ) {
    coord = 1;
  }


  // -- compute area of projection
  const int n = 3;
  for( int i=0; i<n; i++ ) {
    int j = (i < (n - 1)) ? (i + 1) : (0);
    int k = (i > 0) ? (i - 1) : (n - 1);
    switch( coord ) {
    case 0:
      area += (vertex[i].y() * (vertex[j].z() - vertex[k].z()));
      continue;
    case 1:
      area += (vertex[i].x() * (vertex[j].z() - vertex[k].z()));
      continue;
    case 2:
      area += (vertex[i].x() * (vertex[j].y() - vertex[k].y()));
      continue;
    }
  }
    

  // -- scale to get area before projection
  // -- this must be a unit normal
  // -- otherwise, an = normal.length()
  assert( (1.0 - normal.lengthSq()) < EPSILON );
  const Real an = 1.0;
  switch( coord ) {
  case 0:
    area *= (an / (2*ax));
    break;
  case 1:
    area *= (an / (2*ay));
    break;
  case 2:
    area *= (an / (2*az));
  }


  return( area );
}


void ambenv::generate_vertex_normals( Mesh & mesh )
{
  
  std::vector<Vec3> tri_norm( mesh.numTriangles() );
  std::vector<Vec3> vert_norm( mesh.numVertices() );
  {
  for( unsigned i=0; i<mesh.numTriangles(); i++ ) {

    MeshTriangle tri = mesh.getTriangle( i );
    Vec3 a = mesh.getVertex( tri.vert( 0 ) ).v();
    Vec3 b = mesh.getVertex( tri.vert( 1 ) ).v();
    Vec3 c = mesh.getVertex( tri.vert( 2 ) ).v();

    Real BA = (b-a).lengthSq();
    Real CA = (c-a).lengthSq();
    Real CB = (c-b).lengthSq();

    Vec3 ab, ac;

    if( BA <= CA ) {
      if( BA <= CB ) {
	// -- min is BA, origin at C
	ab = (a-c);
	ac = (b-c);
      } else {
	// -- min is CB, origin at A
	ab = (b-a);
	ac = (c-a);
      }
    } else {
      if( CA <= CB ) {
	// -- min is CA, origin at B
	ab = (c-b);
	ac = (a-b);
      } else {
	// -- min is CB, origin at A
	ab = (b-a);
	ac = (c-a);
      }
    }

    tri_norm[i] = ab.cross( ac );
    
    Real magSq = tri_norm[i].lengthSq();
    if( magSq != 0.0 ) {
      tri_norm[i] /= magSq;
    } else {
      // -- points are co-linear
      tri_norm[i] = Vec3( 0.0, 1.0, 0.0 );
    }
    
  }
  }
  {
  for( unsigned i=0; i<mesh.numTriangles(); i++ ) {
    for( unsigned j=0; j<3; j++ ) {
      vert_norm[mesh.getTriangle( i ).vert( j )] += tri_norm[i];
    }
  }
  }
  {
  for( unsigned i=0; i<mesh.numVertices(); i++ ) {
    vert_norm[i].normalize();
    MeshVertex v = mesh.getVertex( i );
    if( vert_norm[i].lengthSq() > 0.0 ) {
      v.n( vert_norm[i] );
    } else {
      v.n( Vec3( 0.0, 1.0, 0.0 ) );
    }
    mesh.setVertex( i, v );
  }
  }

}

void ambenv::doGLMultMatrix( const Matrix4x4 & mat )
{
    static GLfloat xform[16];

    for( unsigned i=0; i<M_DIM; i++ ) {
      for( unsigned j=0; j<M_DIM; j++ ) {
        xform[4*j + i] = mat( i, j );
      }
    }
    glMultMatrixf( xform );
}



